//	GamePlayState.cpp - The main gameplay state.
//  ----------------------------------------------------------------------------
//	This file is part of 'NiallsAVLib', base code for any kind of audiovisual
//	apps.
//	Copyright (C) 2012  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "sound/GenericFileSounder.h"
#include "ClotheslineMissileEntity.h"
#include "PoseidonsPunchingHand.h"
#include "PoseidonHeadEntity.h"
#include "PoseidonChinEntity.h"
#include "BikiniWhaleEntity.h"
#include "PoseidonHeadBack.h"
#include "drawing/Drawer.h"
#include "GamePlayState.h"
#include "ColourHelpers.h"
#include "IslandEntity.h"
#include "OctopusHead.h"
#include "OctopusArm.h"
#include "PlayerShip.h"
#include "Periscope.h"
#include "Scenery.h"
#include "Animal.h"
#include "Player.h"
#include "Random.h"
#include "Wave.h"

using std::wstring;
using std::vector;

//------------------------------------------------------------------------------
GameStateCreator<GamePlayState> GamePlayState::creator;

//------------------------------------------------------------------------------
GamePlayState::GamePlayState(Application& app, SounderPtr& sound):
GameState(app, sound),
moonPos(1180.0f, 32.0f),
moonImage(L"Moon"),
backgroundOffset(0.0f),
playerOnIsland(false),
showJournal(true),
observing(false),
laudanumIndex(0),
alreadyInverted(false),
backgroundColour(0.0f),
backgroundTween(backgroundColour, backgroundColour, 0.1f),
stretchIndex(0.0f),
cycleSky(false),
yScroll(0.0f),
sounder(dynamic_pointer_cast<GenericFileSounder>(sound)),
fadeOut(0.0f)
{
	int i;
	float y;
	TwoFloats tempPos;
	float nextIsland = 1344.0f;
	std::wstring animalNames[] = {L"AnimalLioningo",
								  L"AnimalWheelybox",
								  L"AnimalGirager",
								  L"AnimalSlithersheep",
								  L"AnimalElemu",
								  L"AnimalBeavuck"};

	backgroundTween.setIndex(1.0f);

	animalData[0].name = L"Lioningo";
	animalData[0].image = L"BigAnimalLioningo";
	animalData[0].description = L"The City is alive. At night you can hear it \
breathing. Vast ancient gods stalk its arteries, cloaked in litter and wind. \
No-one sees.\n\
\n\
*I* see.\n\
\n\
The City is alive. And it is waiting.";

	animalData[1].name = L"Locobra";
	animalData[1].image = L"BigAnimalWheelybox";
	animalData[1].description = L"Don't call me. Don't write. Don't email or \
message me. If you see me on the street don't say my name or catch my hand. \
We are strangers now, you & I.\n\
\n\
I love you.";

	animalData[2].name = L"Omercid";
	animalData[2].image = L"BigAnimalGirager";
	animalData[2].description = L"\"How long do we have?\"\n\
\n\
\"Twenty minutes.\"\n\
\n\
\"Right.\" She set the scalpel on the counter and rolled up her sleeve. Slowly, \
carefully, she cut into her arm and started teasing out the fibre optic threads \
woven like a fabric beneath her skin.\n\
\n\
\"Right. Oh fu-\"";

	animalData[3].name = L"Cobrovi";
	animalData[3].image = L"BigAnimalSlithersheep";
	animalData[3].description = L"They start arriving after midnight, chisels \
in hand. Hoodies and scarves covering their faces. Towels to muffle the sound. \
Greetings are exchanged, spots chosen.\n\
\n\
By morning, the once-featureless wall has been transformed into a vast \
sculpture, surreal cartoons clashing with jagged, angular tags carved into the \
stone.";

	animalData[4].name = L"Elemu";
	animalData[4].image = L"BigAnimalElemu";
	animalData[4].description = L"Beyond the city walls the hills hid the \
outside world. At night the city's residents would gather on the rooftops and \
wonder over the glows on the horizon. What fires, what revolutions, were \
responsible for such light in the dark?\n\
\n\
Still the gates remained barred and no figures came striding over the hills.";

	animalData[5].name = L"Duver";
	animalData[5].image = L"BigAnimalBeavuck";
	animalData[5].description = L"";

	stars.push_back(Star(TwoFloats(10.0f), L"Star1"));
	stars.push_back(Star(TwoFloats(32.0f, 128.0f), L"Star2"));
	stars.push_back(Star(TwoFloats(128.0f, 64.0f), L"Star3"));
	stars.push_back(Star(TwoFloats(160.0f, 256.0f), L"Star1"));
	stars.push_back(Star(TwoFloats(192.0f, 32.0f), L"Star2"));
	stars.push_back(Star(TwoFloats(208.0f, 160.0f), L"Star3"));
	stars.push_back(Star(TwoFloats(256.0f, 16.0f), L"Star1"));
	stars.push_back(Star(TwoFloats(384.0f, 192.0f), L"Star2"));
	stars.push_back(Star(TwoFloats(320.0f, 320.0f), L"Star3"));

	stars.push_back(Star(TwoFloats(384.0f+32.0f, 128.0f), L"Star1"));
	stars.push_back(Star(TwoFloats(384.0f+10.0f, 32.0f), L"Star2"));
	stars.push_back(Star(TwoFloats(384.0f+160.0f, 192.0f), L"Star3"));
	stars.push_back(Star(TwoFloats(384.0f+128.0f, 64.0f), L"Star1"));
	stars.push_back(Star(TwoFloats(384.0f+384.0f, 256.0f), L"Star2"));
	stars.push_back(Star(TwoFloats(384.0f+256.0f, 16.0f), L"Star3"));
	stars.push_back(Star(TwoFloats(384.0f+320.0f, 320.0f), L"Star1"));
	stars.push_back(Star(TwoFloats(384.0f+208.0f, 160.0f), L"Star2"));
	stars.push_back(Star(TwoFloats(384.0f+192.0f, 128.0f), L"Star3"));

	stars.push_back(Star(TwoFloats(768.0f+128.0f, 128.0f), L"Star1"));
	stars.push_back(Star(TwoFloats(768.0f+10.0f, 32.0f), L"Star2"));
	stars.push_back(Star(TwoFloats(768.0f+160.0f, 64.0f), L"Star3"));
	stars.push_back(Star(TwoFloats(768.0f+32.0f, 192.0f), L"Star1"));
	stars.push_back(Star(TwoFloats(768.0f+320.0f, 256.0f), L"Star2"));
	stars.push_back(Star(TwoFloats(768.0f+208.0f, 320.0f), L"Star3"));
	stars.push_back(Star(TwoFloats(768.0f+192.0f, 16.0f), L"Star1"));
	stars.push_back(Star(TwoFloats(768.0f+256.0f, 160.0f), L"Star2"));
	stars.push_back(Star(TwoFloats(768.0f+384.0f, 128.0f), L"Star3"));

	stars.push_back(Star(TwoFloats(1152.0f+128.0f, 16.0f), L"Star1"));
	stars.push_back(Star(TwoFloats(1152.0f+32.0f, 320.0f), L"Star2"));
	stars.push_back(Star(TwoFloats(1152.0f+160.0f, 64.0f), L"Star3"));
	stars.push_back(Star(TwoFloats(1152.0f+64.0f, 192.0f), L"Star1"));
	stars.push_back(Star(TwoFloats(1152.0f+320.0f, 32.0f), L"Star2"));
	stars.push_back(Star(TwoFloats(1152.0f+128.0f, 320.0f), L"Star3"));
	stars.push_back(Star(TwoFloats(1152.0f+192.0f, 16.0f), L"Star1"));
	stars.push_back(Star(TwoFloats(1152.0f+256.0f, 160.0f), L"Star2"));
	stars.push_back(Star(TwoFloats(1152.0f+160.0f, 256.0f), L"Star3"));

	stars.push_back(Star(TwoFloats(1536.0f+64.0f, 192.0f), L"Star1"));
	stars.push_back(Star(TwoFloats(1536.0f+256.0f, 160.0f), L"Star2"));
	stars.push_back(Star(TwoFloats(1536.0f+96.0f, 32.0f), L"Star3"));
	stars.push_back(Star(TwoFloats(1536.0f+384.0f, 16.0f), L"Star1"));
	stars.push_back(Star(TwoFloats(1536.0f+16.0f, 64.0f), L"Star2"));
	stars.push_back(Star(TwoFloats(1536.0f+128.0f, 16.0f), L"Star3"));
	stars.push_back(Star(TwoFloats(1536.0f+192.0f, 320.0f), L"Star1"));
	stars.push_back(Star(TwoFloats(1536.0f+256.0f, 128.0f), L"Star2"));
	stars.push_back(Star(TwoFloats(1536.0f+160.0f, 256.0f), L"Star3"));

	//y = 314.0f;
	y = getDrawingAreaSize().y-406.0f;
	for(i=0;i<NumLayers;++i)
	{
		waves.push_back(shared_ptr<Wave>(new Wave(y, ((i%2) > 0))));
		layers[i].addEntity(dynamic_pointer_cast<LayerEntity>(waves.back()));

		//y += ((float)getDrawingAreaSize().y-314.0f)/(float)NumLayers;
		y += 406.0f/(float)NumLayers;
	}

	for(i=0;i<6;++i)
	{
		generateIsland(nextIsland,
					   2 + Random::getGlobal().getRandomInt(3),
					   animalNames[i]);
		observations.push_back(0.0f);

		//nextIsland += 1536.0f + (Random::getGlobal().getRandomFloat() * 512.0f);
		nextIsland += 2048.0f;
	}

	//Bikini whale.
	y = islands[1]->getPosition() - 128.0f;
	bikiniWhale = shared_ptr<BikiniWhaleEntity>(new BikiniWhaleEntity(y));
	layers[2].addEntity(dynamic_pointer_cast<LayerEntity>(bikiniWhale));

	//Poseidon's punching hand.
	y = islands[2]->getPosition() + 1024.0f;
	y += (islands[3]->getPosition() - y) * 0.5f;
	poseidonHand = shared_ptr<PoseidonsPunchingHand>(new PoseidonsPunchingHand(y));
	layers[3].addEntity(dynamic_pointer_cast<LayerEntity>(poseidonHand));

	//Octopus.
	y = islands[3]->getPosition() + 1024.0f;
	y += ((islands[4]->getPosition() - y) * 0.5f) - 160.0f;
	octopusHead = shared_ptr<OctopusHead>(new OctopusHead(y));
	layers[3].addEntity(dynamic_pointer_cast<LayerEntity>(octopusHead));

	tempPos.x = y - 192.0f;
	tempPos.y = 192.0f;
	octopusArms[0] = shared_ptr<OctopusArm>(new OctopusArm(tempPos, 1.0f));
	layers[2].addEntity(dynamic_pointer_cast<LayerEntity>(octopusArms[0]));

	tempPos.x = y - 256.0f;
	tempPos.y = 272.0f;
	octopusArms[1] = shared_ptr<OctopusArm>(new OctopusArm(tempPos, 1.0f));
	layers[4].addEntity(dynamic_pointer_cast<LayerEntity>(octopusArms[1]));

	tempPos.x = y + 192.0f + 96.0f;
	tempPos.y = 208.0f;
	octopusArms[2] = shared_ptr<OctopusArm>(new OctopusArm(tempPos, -1.0f));
	layers[2].addEntity(dynamic_pointer_cast<LayerEntity>(octopusArms[2]));

	tempPos.x = y + 256.0f + 96.0f;
	tempPos.y = 280.0f;
	octopusArms[3] = shared_ptr<OctopusArm>(new OctopusArm(tempPos, -1.0f));
	layers[4].addEntity(dynamic_pointer_cast<LayerEntity>(octopusArms[3]));

	//Periscope.
	y = islands[4]->getPosition() + 1024.0f;
	periscope = shared_ptr<Periscope>(new Periscope(y));
	layers[3].addEntity(dynamic_pointer_cast<LayerEntity>(periscope));
	//Clothesline Missile.
	y = islands[4]->getPosition() + 1536.0f;
	clotheslineMissile = shared_ptr<ClotheslineMissileEntity>(new ClotheslineMissileEntity(y));
	layers[3].addEntity(dynamic_pointer_cast<LayerEntity>(clotheslineMissile));

	//Poseidon's head.
	y = islands[5]->getPosition() + 1536.0f;
	poseidonHead = shared_ptr<PoseidonHeadEntity>(new PoseidonHeadEntity(y));
	layers[3].addEntity(dynamic_pointer_cast<LayerEntity>(poseidonHead));

	poseidonMouth = shared_ptr<PoseidonHeadBack>(new PoseidonHeadBack(y));
	layers[2].addEntity(dynamic_pointer_cast<LayerEntity>(poseidonMouth));

	poseidonChin = shared_ptr<PoseidonChinEntity>(new PoseidonChinEntity(y));
	layers[3].addEntity(dynamic_pointer_cast<LayerEntity>(poseidonChin));

	//Player.
	player = shared_ptr<Player>(new Player());
	layers[3].addEntity(dynamic_pointer_cast<LayerEntity>(player));

	playerShip = shared_ptr<PlayerShip>(new PlayerShip());
	layers[3].addEntity(dynamic_pointer_cast<LayerEntity>(playerShip));
}

//------------------------------------------------------------------------------
GamePlayState::~GamePlayState()
{

}

//------------------------------------------------------------------------------
void GamePlayState::update()
{
	int i;
	float lastLaud = player->getLaudanumTimer();

	if(showJournal)
		return;

	for(i=0;i<(int)waves.size();++i)
		waves[i]->update();

	for(i=0;i<(int)animals.size();++i)
		animals[i]->update();

	playerShip->update();

	player->update();
	if((lastLaud > 0.0f) && (player->getLaudanumTimer() <= 0.0f))
		sounder->stopLoop(laudanumSoundIndex);
	if(player->getHunger() >= 1.0f)
	{
		if(player->getLives() > 0)
			sounder->triggerSound(L"SFXLoseHeart");
		player->decrementLife();
	}
	if(!playerOnIsland)
	{
		player->setPosition(playerShip->getPosition()-48.0f);
		shipCollisions();
		bikiniWhaleCollisions();
		octopusCollisions();
	}
	else
	{
		if(player->getPosition() < (islands[islandIndex]->getPosition()+128.0f))
		{
			player->setOnIsland(false);
			playerOnIsland = false;
			observing = false;
			player->setWriting(false);
			playerShip->setPosition(islands[islandIndex]->getPosition() - 16.0f);
			//sounder->stopLoop(writingIndex);
		}
	}

	if(observing)
	{
		observations[islandIndex] += 0.005f;
		if(observations[islandIndex] >= 1.0f)
		{
			observations[islandIndex] = 1.0f;
			observing = false;
			player->setWriting(false);

			journal.addPage(animalData[islandIndex].name,
							animalData[islandIndex].image,
							animalData[islandIndex].description);
			showJournal = true;
			sounder->stopLoop(writingIndex);
		}
	}

	if(!backgroundTween.getFinished())
		backgroundTween.tick();
	else if(observing)
	{
		ThreeFloats tempCol;

		tempCol.x = Random::getGlobal().getRandomFloat();
		tempCol.y = 0.75f;
		tempCol.z = 0.66f;
		tempCol = ColourHelpers::hslToRgb(tempCol);
		backgroundTween.reset(tempCol);
	}
	/*else if((player->getLaudanumTimer() > 0.0f) && (laudanumIndex == 0))
	{
		ThreeFloats tempCol;

		tempCol.x = Random::getGlobal().getRandomFloat();
		tempCol.y = 0.75f;
		tempCol.z = 0.66f;
		tempCol = ColourHelpers::hslToRgb(tempCol);
		backgroundTween.reset(tempCol);
	}*/
	else if(backgroundColour != 0.0f)
		backgroundTween.reset(0.0f);

	if((player->getLaudanumTimer() > 0.0f) &&
	   ((laudanumIndex == 2) || (laudanumIndex == 3)))
	{
		if(laudanumIndex == 2)
			stretchIndex += 0.01f;
		else
			stretchIndex += 0.005f;
		if(stretchIndex > 1.0f)
			stretchIndex = 0.0f;
	}

	for(i=0;i<NumOctopusArms;++i)
		octopusArms[i]->update();

	if(!bikiniWhale->getActivated())
	{
		if(playerShip->getPosition() > (bikiniWhale->getPosition() - 768.0f))
			bikiniWhale->activate();
	}
	bikiniWhale->update();

	if(!pacman.getActive())
	{
		if(playerShip->getPosition() > (islands[2]->getPosition() - 1024.0f))
			pacman.activate();
	}
	else
	{
		const TwoFloats& pacmanPos = pacman.getPosition();

		if(moonImage == L"Moon")
		{
			if((pacmanPos.x+100.0f) >= moonPos.x)
				moonImage = L"MoonEaten";
		}

		for(i=(stars.size()-1);i>=0;--i)
		{
			if(stars[i].position.getDistance(pacmanPos) < 100.0f)
				stars.erase(stars.begin()+i);
		}
	}
	pacman.update();

	if(!periscope->getActive())
	{
		if(playerShip->getPosition() > (periscope->getPosition() - 512.0f))
			periscope->activate();
	}
	periscope->update();

	if(!clotheslineMissile->getActivated() && periscope->getFinished())
		clotheslineMissile->activate();
	clotheslineMissile->update();

	if(!poseidonHand->getActivated())
	{
		if(playerShip->getPosition() > (poseidonHand->getPosition() - 149.0f))
		{
			poseidonHand->activate();

			if(player->getLayer() <= 3)
			{
				playerShip->punch();
				player->decrementLife();
				sounder->triggerSound(L"SFXPunch");
				sounder->triggerSound(L"SFXLoseHeart");
			}
		}
	}
	else
	{
		poseidonHand->update();

		if(!playerShip->isPunched() && (player->getLayer() <= 3))
		{
			float shipPos = playerShip->getPosition()+32.0f;

			if(((shipPos + (224.0f-32.0f)) > (poseidonHand->getPosition()+32.0f)) &&
				(shipPos < (poseidonHand->getPosition() + (266.0f-32.0f))))
			{
				playerShip->punch();
				player->decrementLife();
				sounder->triggerSound(L"SFXPunch");
				sounder->triggerSound(L"SFXLoseHeart");
			}
		}
	}

	if(playerShip->isPunched() ||
	   playerShip->isOctoPunched() ||
	   playerShip->getSinking())
	{
		player->setPosition(TwoFloats(playerShip->getPosition()-48.0f,
									  playerShip->getYPos()+8.0f));
	}

	if(playerShip->getPosition() > (poseidonHead->getPosition() - 460.0f))
	{
		//cycleSky = true;

		if(playerShip->getLayer() != 2)
		{
			layers[playerShip->getLayer()].removeEntity(dynamic_pointer_cast<LayerEntity>(player));
			layers[playerShip->getLayer()].removeEntity(dynamic_pointer_cast<LayerEntity>(playerShip));
			player->setLayer(2, getDrawingAreaSize().y);
			playerShip->setLayer(2, getDrawingAreaSize().y);
			layers[2].addEntity(dynamic_pointer_cast<LayerEntity>(player));
			layers[2].addEntity(dynamic_pointer_cast<LayerEntity>(playerShip));
		}
		playerShip->goIntoMouth();
	}

	if((yScroll < 0.0f) && (yScroll > -2048.0f))
		yScroll -= 1.0f;
}

//------------------------------------------------------------------------------
void GamePlayState::draw(DrawerPtr& d, float timestepInterp)
{
	int i;
	float playerPos;
	TwoFloats tempPos;
	const TwoFloats centre(d->getDrawingAreaSize() * 0.5f);

	if(playerOnIsland)
		playerPos = player->getPosition();
	else
		playerPos = playerShip->getPosition();

	if(playerPos < centre.x)
		backgroundOffset = 0.0f;
	else if((playerPos-centre.x) > (poseidonHead->getPosition() - 516.0f))
		backgroundOffset = (poseidonHead->getPosition() - 516.0f)/8.0f;
	else
		backgroundOffset = (playerPos - centre.x)/8.0f;

	//Scroll everything up if the player's been sunk by the whale.
	if(yScroll < 0.0f)
		d->translate(TwoFloats(0.0f, yScroll));

	//Fill background colour.
	d->fillRect(0.0f, d->getDrawingAreaSize(), backgroundColour);
	if(yScroll < 0.0f)
	{
		d->fillGradient(TwoFloats(0.0f, d->getDrawingAreaSize().y),
						TwoFloats(d->getDrawingAreaSize().x, 2048.0f),
						ThreeFloats(0.25f, 0.25f, 0.75f),
						ThreeFloats(0.0f, 0.0f, 0.05f));
	}

	if(!showJournal)
	{
		if((player->getLaudanumTimer() > 0.0f) && (laudanumIndex == 1))
		{
			d->scale(TwoFloats(1.0f, -1.0f));
			d->translate(TwoFloats(0.0f, -d->getDrawingAreaSize().y));
		}
		else if((player->getLaudanumTimer() > 0.0f) &&
				((laudanumIndex == 2) || (laudanumIndex == 3)))
		{
			TwoFloats tempStretch;
			TwoFloats tempPos;

			tempStretch.x = 1.1f + (sinf(stretchIndex * 2.0f * NIALL_PI)*0.1f);
			tempStretch.y = 1.1f - (cosf(stretchIndex * 2.0f * NIALL_PI)*0.1f);
			d->scale(tempStretch);
			tempPos = ((d->getDrawingAreaSize() * tempStretch) - d->getDrawingAreaSize()) * 0.5f;
			tempPos.x = -tempPos.x;
			tempPos.y = -tempPos.y;
			d->translate(tempPos);
		}
	}

	//Draw stars.
	for(i=0;i<(int)stars.size();++i)
	{
		tempPos = stars[i].position;
		tempPos.x -= backgroundOffset*0.5f;
		if(tempPos.x < d->getDrawingAreaSize().x)
			d->drawImage(stars[i].image, tempPos);
	}

	//Draw moon.
	tempPos = moonPos;
	tempPos.x -= backgroundOffset*0.5f;
	d->drawImage(moonImage, tempPos);

	//Draw pacman.
	pacman.draw(d, backgroundOffset*0.5f);

	//Draw layers.
	for(i=0;i<NumLayers;++i)
	{
		if((yScroll < -0.0f) && (i == (NumLayers-1)))
		{
			d->fillRect(TwoFloats(0.0f, d->getDrawingAreaSize().y),
						TwoFloats(d->getDrawingAreaSize().x, 96.0f),
						1.0f);
			d->fillGradient(TwoFloats(0.0f, d->getDrawingAreaSize().y+96.0f),
							TwoFloats(d->getDrawingAreaSize().x, 720.0f),
							1.0f,
							1.0f,
							1.0f,
							0.0f);
		}
		layers[i].draw(d, backgroundOffset * 8.0f);
	}

	//Draw hud.
	/*d->drawImage(L"NauseaText", 10.0f);
	d->drawImage(L"HungerText", TwoFloats(10.0f, 70.0f));

	d->drawSubImage(L"NauseaBar",
					TwoFloats(212.0f, 10.0f),
					0.0f,
					TwoFloats(player->getNausea(), 1.0f));
	d->drawSubImage(L"HungerBar",
					TwoFloats(212.0f, 70.0f),
					0.0f,
					TwoFloats(player->getHunger(), 1.0f));*/

	//Reset translation for hud stuff.
	if(yScroll < 0.0f)
		d->translate(TwoFloats(0.0f, -yScroll));

	if(player->getLives() > 0)
		d->drawImage(L"Heart", TwoFloats(d->getDrawingAreaSize().x-60.0f, 16.0f));
	if(player->getLives() > 1)
		d->drawImage(L"Heart", TwoFloats(d->getDrawingAreaSize().x-(60.0f*2.0f), 16.0f));
	if(player->getLives() > 2)
		d->drawImage(L"Heart", TwoFloats(d->getDrawingAreaSize().x-(60.0f*3.0f), 16.0f));

	//Draw wee journal.
	if(playerOnIsland)
	{
		d->drawImage(L"WeeJournalOpen",
					 TwoFloats(d->getDrawingAreaSize().x - 100.0f,
							   d->getDrawingAreaSize().y - 74.0f));
		d->drawSubImage(L"WeeJournalText",
						TwoFloats(d->getDrawingAreaSize().x - 100.0f,
								  d->getDrawingAreaSize().y - 74.0f),
						TwoFloats(0.0f, 0.0f),
						TwoFloats(1.0f, observations[islandIndex]));
	}
	else
	{
		d->drawImage(L"WeeJournalClosed",
					 TwoFloats(d->getDrawingAreaSize().x - 62.0f,
							   d->getDrawingAreaSize().y - 74.0f));
	}

	if(showJournal)
		journal.draw(d);
	else if((player->getLaudanumTimer() > 0.0f) && (laudanumIndex == 3))
		whirlingAnimals.draw(d);

	if(fadeOut > 0.0f)
	{
		d->fillRect(0.0f, d->getDrawingAreaSize(), 0.0f, fadeOut);
		if(fadeOut < 1.0f)
			fadeOut += 0.05f;
	}
}

//------------------------------------------------------------------------------
void GamePlayState::keyDown(const wstring& action,
							const wstring& keyPressed,
							int keyModifier)
{
	ThreeFloats tempCol;
	int currentLayer = playerShip->getLayer();

	/*if(keyPressed == L"f")
		fadeOut += 0.05f;
	else */if(action == L"Escape")
	{
		if(showJournal)
			showJournal = false;
		else
			nextState(L"TitleScreen");
	}
	else if(action == L"Journal")
		showJournal = !showJournal;
	else if(showJournal)
	{
		if(action == L"P1Left")
			journal.changePage(-1);
		else if(action == L"P1Right")
			journal.changePage(1);
	}
	/*else if(action == L"Laudanum")
	{
		if(player->getLaudanumTimer() <= 0.0f)
		{
			player->decreaseNausea(0.2f);

			laudanumIndex = Random::getGlobal().getRandomInt(4);
			while(alreadyInverted && (laudanumIndex == 1))
				laudanumIndex = Random::getGlobal().getRandomInt(4);
			if(laudanumIndex == 0)
			{
				tempCol.x = Random::getGlobal().getRandomFloat();
				tempCol.y = 0.75f;
				tempCol.z = 0.66f;
				tempCol = ColourHelpers::hslToRgb(tempCol);
				backgroundTween.reset(tempCol);
			}
			else if(laudanumIndex == 1)
				alreadyInverted = true;

			sounder->triggerSound(L"SFXLaudanum");
			laudanumSoundIndex = sounder->triggerLoop(L"LaudanumDrone");
		}
	}*/
	else if(playerOnIsland)
	{
		if(!observing)
		{
			if(action == L"P1Left")
				player->move(-1);
			else if(action == L"P1Right")
				player->move(1);
			else if(action == L"Action")
			{
				if(observations[islandIndex] < 1.0f)
				{
					//observing = true;
					observing = !observing;
					if(observing)
					{
						writingIndex = sounder->triggerLoop(L"SFXWriting");
						player->move(0);
						player->setWriting(true);
					}
				}
			}
		}
	}
	else if(action == L"P1Up")
	{
		if((playerShip->getLayer() > 0) && (isLayerFree(currentLayer-1)))
		{
			layers[currentLayer].removeEntity(dynamic_pointer_cast<LayerEntity>(player));
			layers[currentLayer].removeEntity(dynamic_pointer_cast<LayerEntity>(playerShip));
			--currentLayer;
			player->setLayer(currentLayer, getDrawingAreaSize().y);
			playerShip->setLayer(currentLayer, getDrawingAreaSize().y);
			layers[currentLayer].addEntity(dynamic_pointer_cast<LayerEntity>(player));
			layers[currentLayer].addEntity(dynamic_pointer_cast<LayerEntity>(playerShip));
		}
	}
	else if(action == L"P1Down")
	{
		if((playerShip->getLayer() < 6) && (isLayerFree(currentLayer+1)))
		{
			layers[currentLayer].removeEntity(dynamic_pointer_cast<LayerEntity>(player));
			layers[currentLayer].removeEntity(dynamic_pointer_cast<LayerEntity>(playerShip));
			++currentLayer;
			player->setLayer(currentLayer, getDrawingAreaSize().y);
			playerShip->setLayer(currentLayer, getDrawingAreaSize().y);
			layers[currentLayer].addEntity(dynamic_pointer_cast<LayerEntity>(player));
			layers[currentLayer].addEntity(dynamic_pointer_cast<LayerEntity>(playerShip));
		}
	}
	else if(action == L"P1Right")
		playerShip->move(1);
	else if(action == L"P1Left")
		playerShip->move(-1);
	/*else if(action == L"SeaWater")
	{
		if(player->getNausea() < 0.66f)
		{
			sounder->triggerSound(L"SFXDrink");

			player->decreaseHunger(0.125f);
			player->increaseNausea(0.33f);
			if(player->getNausea() >= 1.0f)
			{
				player->decrementLife();
				sounder->triggerSound(L"SFXLoseHeart");
			}
		}
	}*/
}

//------------------------------------------------------------------------------
void GamePlayState::keyUp(const wstring& action,
							const wstring& keyPressed,
							int keyModifier)
{
	if(playerOnIsland)
	{
		if((action == L"P1Right") || (action == L"P1Left"))
			player->move(0);
		/*else if(action == L"Action")
		{
			observing = false;
			sounder->stopLoop(writingIndex);
		}*/
	}
	else if((action == L"P1Right") || (action == L"P1Left"))
		playerShip->move(0);
}

//------------------------------------------------------------------------------
void GamePlayState::generateIsland(float xPos,
								   int layer,
								   const std::wstring& animalName)
{
	int i;
	int tempint;
	TwoFloats tempPos;
	TwoFloats sceneryPos;
	//std::wstring images[] = {L"Tree", L"Rock", L"Bush"};
	std::vector<std::wstring> images;

	images.push_back(L"Tree");
	images.push_back(L"Tree");
	images.push_back(L"Rock");
	images.push_back(L"Rock");
	images.push_back(L"Bush");
	images.push_back(L"Bush");

	tempPos.x = xPos;
	tempPos.y = 314.0f + ((((float)getDrawingAreaSize().y-314.0f)/(float)NumLayers) * (float)layer);
	tempPos.y -= 64.0f;
	islands.push_back(shared_ptr<IslandEntity>(new IslandEntity(tempPos)));
	layers[layer].addEntity(dynamic_pointer_cast<LayerEntity>(islands.back()));

	for(i=0;i<4;++i)
	{
		animals.push_back(shared_ptr<Animal>(new Animal(animalName,
														xPos+128.0f,
														tempPos.y)));
		layers[layer].addEntity(dynamic_pointer_cast<LayerEntity>(animals.back()));
	}

	sceneryPos.x = tempPos.x + 360.0f;
	for(i=0;i<4;++i)
	{
		tempint = Random::getGlobal().getRandomInt(images.size());
		if(images[tempint] == L"Tree")
		{
			sceneryPos.x -= 180.0f;
			sceneryPos.y = tempPos.y - (384.0f-72.0f);
		}
		else if(images[tempint] == L"Rock")
		{
			sceneryPos.x -= 48.0f;
			sceneryPos.y = tempPos.y + 0.0f;
		}
		else if(images[tempint] == L"Bush")
		{
			sceneryPos.x -= 48.0f;
			sceneryPos.y = tempPos.y - 12.0f;
		}
		sceneryPos.y -= Random::getGlobal().getRandomFloat() * 8.0f;
		scenery.push_back(shared_ptr<Scenery>(new Scenery(sceneryPos, images[tempint])));
		layers[layer+1].addEntity(dynamic_pointer_cast<LayerEntity>(scenery.back()));

		if(images[tempint] == L"Tree")
			sceneryPos.x += 180.0f;
		else
			sceneryPos.x += 48.0f;

		images.erase(images.begin()+tempint);

		sceneryPos.x +=	(360.0f/3.0f);
		sceneryPos.x += (Random::getGlobal().getRandomFloat() * 64.0f) - 32.0f;
	}
}

//------------------------------------------------------------------------------
void GamePlayState::shipCollisions()
{
	int i;
	unsigned int j;
	TwoFloats tempPos;
	const float playerPos = playerShip->getPosition();
	const int numLayerEntities = layers[playerShip->getLayer()].getNumEntities();

	for(i=0;i<numLayerEntities;++i)
	{
		const shared_ptr<LayerEntity>& entity = layers[playerShip->getLayer()].getEntity(i);

		if(entity->getType() == L"Island")
		{
			if((playerPos < (entity->getPosition() + 256.0f + 512.0f)) &&
			   ((playerPos + 256.0f) > (entity->getPosition() + 256.0f)))
			{
				//Stop ship moving.
				playerShip->move(0);

				player->setOnIsland(true);
				playerOnIsland = true;
				for(j=0;j<islands.size();++j)
				{
					if(dynamic_pointer_cast<IslandEntity>(entity) == islands[j])
					{
						islandIndex = j;
						break;
					}
				}
				tempPos = entity->getPosition();
				tempPos.x += 128.0f;
				tempPos.y -= 128.0f;
				player->setPosition(tempPos.x);
				player->setIslandPos(entity->getPosition() + 128.0f);

				break;
			}
		}
	}
}

//------------------------------------------------------------------------------
void GamePlayState::octopusCollisions()
{
	const int layer = player->getLayer();
	const float playerPos = player->getPosition();

	if(layer == 2)
	{
		if((playerPos > (octopusArms[0]->getPosition() - 192.0f)) &&
		   (playerPos < (octopusArms[0]->getPosition() + 256.0f)))
		{
			octopusArms[0]->punch();
			if(!playerShip->isOctoPunched() && !octopusArms[0]->isPunching())
			{
				playerShip->octopusPunch();
				player->decrementLife();
				sounder->triggerSound(L"SFXPunch");
				sounder->triggerSound(L"SFXLoseHeart");
			}
		}
	}
	else if(layer == 4)
	{
		if((playerPos > (octopusArms[1]->getPosition() - 192.0f)) &&
		   (playerPos < (octopusArms[1]->getPosition() + 256.0f)))
		{
			octopusArms[1]->punch();
			if(!playerShip->isOctoPunched() && !octopusArms[1]->isPunching())
			{
				playerShip->octopusPunch();
				player->decrementLife();
				sounder->triggerSound(L"SFXPunch");
				sounder->triggerSound(L"SFXLoseHeart");
			}
		}
	}
}

//------------------------------------------------------------------------------
void GamePlayState::bikiniWhaleCollisions()
{
	const float playerPos = player->getPosition();
	const float whalePos = bikiniWhale->getPosition();

	if(playerShip->getSinking())
		return;

	if(player->getLayer() == 2)
	{
		if(whalePos < 2661.0f)
		{
			if(fabsf(whalePos-playerPos) < 32.0f)
			{
				playerShip->sink();
				yScroll -= 1.0f;
			}
		}
	}
}

//------------------------------------------------------------------------------
bool GamePlayState::isLayerFree(int layer)
{
	int i;
	bool retval = true;
	const float playerPos = playerShip->getPosition();
	const int numLayerEntities = layers[layer].getNumEntities();

	for(i=0;i<numLayerEntities;++i)
	{
		const shared_ptr<LayerEntity>& entity = layers[layer].getEntity(i);

		if(entity->getType() == L"Island")
		{
			if((playerPos < (entity->getPosition() + 256.0f + 512.0f)) &&
			   ((playerPos + 256.0f) > (entity->getPosition() + 256.0f)))
			{
				retval = false;

				break;
			}
		}
	}

	return retval;
}
