//	GamePlayState.h - The main gameplay state.
//  ----------------------------------------------------------------------------
//	This file is part of 'NiallsAVLib', base code for any kind of audiovisual
//	apps.
//	Copyright (C) 2012  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef GAMEPLAYSTATE_H_
#define GAMEPLAYSTATE_H_

#include "WhirlingAnimals.h"
#include "JournalDisplay.h"
#include "GameState.h"
#include "TwoFloats.h"
#include "Tweener.h"
#include "Pacman.h"
#include "Layer.h"

#include <vector>

class ClotheslineMissileEntity;
class PoseidonsPunchingHand;
class PoseidonHeadEntity;
class PoseidonChinEntity;
class GenericFileSounder;
class BikiniWhaleEntity;
class PoseidonHeadBack;
class IslandEntity;
class OctopusHead;
class OctopusArm;
class PlayerShip;
class Periscope;
class Scenery;
class Animal;
class Player;
class Wave;

///	The main gameplay state.
class GamePlayState : public GameState
{
  public:
	///	Constructor.
	GamePlayState(Application& app, SounderPtr& sound);
	///	Destructor.
	~GamePlayState();

	///	Called frequently to update the state.
	void update();
	///	Called to draw the state.
	void draw(DrawerPtr& d, float timestepInterp);

	///	Called when a key down event is received.
	void keyDown(const std::wstring& action,
				 const std::wstring& keyPressed,
				 int keyModifier);
	///	Called when a key up event is received.
	void keyUp(const std::wstring& action,
			   const std::wstring& keyPressed,
			   int keyModifier);

	///	Returns the name of this state.
	const std::wstring getName() const {return L"GamePlayState";};
	///	Used to construct instances of this state.
	static GameStateCreator<GamePlayState> creator;
  private:
	///	Helper method. Generates an island.
	void generateIsland(float xPos, int layer, const std::wstring& animalName);
	///	Helper method. Tests if it's possible for the ship to switch layers.
	bool isLayerFree(int layer);
	///	Helper method. Handles PlayerShip collisions.
	void shipCollisions();
	///	Helper method. Handles PlayerShip->Octopus collisions.
	void octopusCollisions();
	///	Helper method. Handles PlayerShip->BikiniWhale collisions.
	void bikiniWhaleCollisions();

	///	The data for each animal.
	struct AnimalData
	{
		///	The animal's name.
		std::wstring name;
		///	The animal's image.
		std::wstring image;
		///	The animal's description.
		std::wstring description;
	};

	///	Simple struct representing a star.
	struct Star
	{
		///	The star's position.
		TwoFloats position;
		///	The star's image.
		std::wstring image;

		///	Constructor.
		Star(const TwoFloats& pos, const std::wstring& im):
		position(pos),
		image(im)
		{

		};
	};

	///	Some constants.
	enum
	{
		NumLayers = 8,
		NumOctopusArms = 4
	};

	///	The stars' positions.
	std::vector<Star> stars;
	///	The moon's position.
	TwoFloats moonPos;
	///	The moon's image.
	std::wstring moonImage;
	///	How much to offset the stars and the moon along the x-axis.
	float backgroundOffset;
	///	The flying pacman.
	Pacman pacman;

	///	The foreground layers.
	Layer layers[NumLayers];

	///	The waves.
	std::vector<shared_ptr<Wave> > waves;
	///	The islands.
	std::vector<shared_ptr<IslandEntity> > islands;
	///	The island scenery.
	std::vector<shared_ptr<Scenery> > scenery;

	///	The animals.
	std::vector<shared_ptr<Animal> > animals;

	///	The player's ship.
	shared_ptr<PlayerShip> playerShip;
	///	The player.
	shared_ptr<Player> player;
	///	True if the player is on an island.
	bool playerOnIsland;
	///	The index of the island the player is on.
	int islandIndex;

	///	The bikini whale.
	shared_ptr<BikiniWhaleEntity> bikiniWhale;

	///	The submarine periscope.
	shared_ptr<Periscope> periscope;
	///	The clothesline missile.
	shared_ptr<ClotheslineMissileEntity> clotheslineMissile;

	///	Poseidon's punching hand.
	shared_ptr<PoseidonsPunchingHand> poseidonHand;
	///	The octopus head.
	shared_ptr<OctopusHead> octopusHead;
	///	The octopus arms.
	shared_ptr<OctopusArm> octopusArms[NumOctopusArms];

	///	Poseidon's head.
	shared_ptr<PoseidonHeadEntity> poseidonHead;
	///	Poseidon's mouth.
	shared_ptr<PoseidonHeadBack> poseidonMouth;
	///	Poseidon's chin.
	shared_ptr<PoseidonChinEntity> poseidonChin;

	///	Arwin's journal.
	JournalDisplay journal;
	///	Whether to show the journal or not.
	bool showJournal;
	///	How much the player has observed on each island.
	std::vector<float> observations;
	///	True if the player is observing.
	bool observing;
	///	The data for the various animals.
	AnimalData animalData[6];

	///	Index of which laudanum effect we're currently displaying.
	int laudanumIndex;
	///	Used to ensure the upside down effect only happens once.
	bool alreadyInverted;

	///	The background colour.
	ThreeFloats backgroundColour;
	///	The background colour tweener.
	STweener<ThreeFloats> backgroundTween;
	///	Used to stretch and squeeze the screen.
	float stretchIndex;
	///	The whirling animals laudanum effect.
	WhirlingAnimals whirlingAnimals;

	///	True if we're cycling the sky colour.
	bool cycleSky;

	///	Used to scroll everything up when the player's ship is sunk.
	float yScroll;

	///	Our sounder.
	shared_ptr<GenericFileSounder> sounder;
	///	The writing sound index.
	int writingIndex;
	///	The laudanum sound index.
	int laudanumSoundIndex;

	///	Temp.
	float fadeOut;
};

#endif
