//	Player.h - The Arwin character.
//  ----------------------------------------------------------------------------
//	This file is part of 'NiallsAVLib', base code for any kind of audiovisual
//	apps.
//	Copyright (C) 2012  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef PLAYER_H_
#define PLAYER_H_

#include "c0xHeaders.h"
#include "TwoFloats.h"
#include "Layer.h"

class Drawer;

///	The Arwin character.
class Player : public LayerEntity
{
  public:
	///	Constructor.
	Player();
	///	Destructor.
	~Player();

	///	Moves the player.
	void update();
	///	Draws the player.
	void draw(shared_ptr<Drawer>& d, float xOffset);

	///	Returns the entity's position on the x-axis.
	float getPosition() const {return position.x;};
	///	Returns true if the entity should be considered for collisions.
	bool getCollides() const {return true;};
	///	Returns the entity's type name.
	const std::wstring getType() const {return L"Player";};

	///	Sets the player's x-axis position (when it's being controlled by the ship).
	void setPosition(float pos);
	///	Sets the player's position (when it's just landed on an island).
	void setPosition(const TwoFloats& pos);
	///	Tells the player to move (-1 = left, 0 = stop, 1 = right).
	void move(int val);
	///	Tells the player the start position of the current island.
	void setIslandPos(float pos);
	///	So we know when the player's on an island.
	void setOnIsland(bool val);

	///	Called when the player is writing an observation.
	void setWriting(bool val);

	///	Moves the player to the set layer.
	void setLayer(int val, float screenHeight);
	///	Returns the player's current layer.
	int getLayer() const {return layer;};

	///	Returns the player's nausea level.
	float getNausea() const {return nausea;};
	///	Returns the player's hunger level.
	float getHunger() const {return hunger;};
	///	Returns the player's number of lives.
	int getLives() const {return lives;};
	///	Returns the player's laudanum timer.
	float getLaudanumTimer() const {return laudanumTimer;};

	///	Increases the player's nausea level.
	void increaseNausea(float val);
	///	Decreases the player's nausea level.
	void decreaseNausea(float val);
	///	Decreases the player's hunger level.
	void decreaseHunger(float val);
	///	Decrements a life (also reduces hunger).
	void decrementLife();
  private:
	///	The player's position.
	TwoFloats position;
	///	The player's current layer.
	int layer;

	///	Used to bounce the player while he's walking.
	float bounceIndex;
	///	Whether the player's moving (on land) or not.
	int moving;

	///	Used to draw the player facing the correct direction.
	float dirScale;

	///	Used to determine how far the player can move on the island.
	float islandPos;
	///	Whether or not the player's currently on an island.
	bool onIsland;

	///	True if the player is writing an observation.
	bool writing;

	///	The player's current nausea level.
	float nausea;
	///	The player's current hunger level.
	float hunger;
	///	How many lives the player still has.
	int lives;

	///	Used to control how frequently the player can take the laudanum.
	float laudanumTimer;
};

#endif
